package org.mvel2.tests.core.res;


import java.util.HashMap;
import java.util.Map;

import org.mvel2.templates.TemplateRuntime;

public class TestMVEL197 {

  private String name1;

  private String name2;

  public String getName1() {
    return name1;
  }

  public void setName1(String name1) {
    this.name1 = name1;
  }

  public String getName2() {
    return name2;
  }

  public void setName2(String name2) {
    this.name2 = name2;
  }
}