package org.mvel2.tests.perftests;

import java.io.IOException;
import java.io.OutputStream;

public class NullOutputStream extends OutputStream {
  public void write(int i) throws IOException {

  }
}
