package org.mvel2.tests.core.res;

/**
 * Created by IntelliJ IDEA.
 * User: brockm
 * Date: Jun 22, 2007
 * Time: 6:31:04 PM
 * To change this template use File | Settings | File Templates.
 */
public interface TestInterface {
  public String getName();

  public boolean isFoo();
}
