package org.mvel2.tests.core.res;

public class ScenarioType {
  public static class Set {
    public static final ScenarioType ADD = new ScenarioType();
    public static final ScenarioType EDIT = new ScenarioType();
  }
}