package org.mvel2.tests.core.res;

import java.util.Set;


public class Cake {
  private Set ingredients;

  public Set getIngredients() {
    return ingredients;
  }

  public void setIngredients(Set ingredients) {
    this.ingredients = ingredients;
  }
}

